// gcard_example01.cpp

//A program that takes a sequence from the input file and doubles each of its elements.
// compile with
// c++ -o gcard_ex01 gcard_example01.cpp -lOpenCL -std=c++11
// execute with
// ./gcard_ex01

#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <iostream>
#include <fstream>


#include "GraphicCardOpenCL.cpp"
#include "generatingSequenceFromFile.cpp"




int main(){

    myint *aRAM;
    myint nRAM;

    readSequenceFromFile("input01.txt", &aRAM, &nRAM );



    // Allocate an object myGPU that is of the type GraphicCard
    // myGPU object will contain the kernel written in parallel01.cl
    GraphicCard myGPU("parallel01.cl");

    // We will place the sequence aRAM to the device memory.
    // The kernel will look for a sequence "a" on the device memory.
    // We will create this sequence by copying the sequence aRAM from the host memory.
    myGPU.writeDeviceMemory("a",aRAM,nRAM);
    // Similarly, we need to place the variable "n" on the graphic card.
    // Since the only allowed objects on the graphic cards are sequences, and n is
    // a single element, it must be treated as an one-element sequence.
    // Luckily, in C++ every pointer is an one-element sequence, and so is &nRAM.
    myGPU.writeDeviceMemory("n",&nRAM,1);


    // The GPU is given instruction to execute the kernel on at least nRAM processing elements.
    // Caution: GPU may actually invoke more than nRAM elements, and you can't control this.
    //          Hardware design of GPU organizes the processing elements in groups and only the entire
    //          group can be called.
    myGPU.executeKernel("eachTermGetsDoubled",nRAM);



    // After the kernel execution we need to retrieve the new modified sequence "a".
    // The sequence has to be copied from the device memory back to the RAM memory.
    // We can accomplish this by first creating the space in RAM sufficient to store the new sequence.
    myint * resultRAM;
    resultRAM=new myint[nRAM];

    // Now since the space is created, we call the function that reads the content of the
    // device memory and stores it in the sequence resultRAM.
    myGPU.readDeviceMemory("a",resultRAM,nRAM);


    //The final step is to output the result in the file output.txt
    std::ofstream fOutput;
    fOutput.open("output01.txt");

    myint mLen=nRAM;
    if(mLen>30){mLen=30;}
    for(myint i=0;i<mLen;i++){
        fOutput<<resultRAM[i]<<" ";
    }
    fOutput<<std::endl;

    fOutput.close();
    delete [] aRAM;
    delete [] resultRAM;



    return 1;
}
